"""email_responder: Log into an email and send emails.

Example:

    # Initialize the email responder
    responder = EmailResponder(example@email.com, badpassword)
    # Login
    responder.login()
    # The recipient for the email
    recipient = "other@email.com"
    # The subject of the emails
    subject = "This is the subject"
    # Plain text version of the email
    text = "This is the body of the email."
    # HTML version of the email
    html = '''\
    <html>
      <body>
        <img src="cid:image0"/>
        <br>
        <p>This is the body of the email.</p>
        <br>
        <img src="cid:image1"/>
      </body>
    </html>'''
    # List of images
    images = ["./images/example1.png", "./images/example2.png"]
    # List of attachments
    attachments = ["./some_file.txt", "./folder/another_file.pdf"]

    # Sends a plain text email
    responder.send_singlepart_msg(recipient, subject, text)
    # Sends a multipart (MIME) email with only plain text
    responder.send_multipart_msg(recipient, subject, text)
    # Sends a multipart (MIME) email with all options
    responder.send_multipart_msg(recipient, subject, text, html, images, attachments)

"""

# Imports from other packages
import os
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib, ssl


class EmailResponder:
    """EmailResponder object for sending emails.

    Attributes:
        email (str): The email to send emails from.
        app_password (str): The password for the email.
        server (SMTP_SSL): The SMTP server to use. Defaults to None.

    """

    def __init__(self, email, app_password):
        """Initialize an EmailResponder instance.

        Args:
            email (str): The email to send emails from.
            app_password (str): The password for the email.

        """

        self.email = email
        self.app_password = app_password
        self.server = None


    def login(self, host="smtp.gmail.com", port=465):
        """Logs in the EmailResponder to the SMTP server.

        Args:
            host (str): The smtp host to log into. Default host is Gmail.
            port (int): The port number to connect through. Default port is
                465, which is needed for a secure SSL connection.

        Returns:
            None

        """

        context = ssl.create_default_context()
        self.server = smtplib.SMTP_SSL(host, port, context=context)
        self.server.login(self.email, self.app_password)


    def logout(self):
        """Logs out the EmailResponder from the SMTP server.

        Args:
            None

        Returns:
            None

        """

        self.server.quit()
        self.server = None


    def send_singlepart_msg(self, recipient, subject, message):
        """Sends a plain text email to the recipient.

        Args:
            recipient (str): The email address to send the email to.
            subject (str): The subject of the email.
            message (str): The message of the email.

        Returns:
            None

        """

        msg = "Subject: {}\n\n{}".format(subject, message)
        self.server.sendmail(self.email, recipient, msg)


    def send_multipart_msg(self, recipient, subject, text, html=None, images=None,
                            attachments=None):
        """Sends a multipart (MIME) email to the recipient.

        Args:
            recipient (str): The email address to send the email to.
            subject (str): The subject of the email.
            text (str): The plain text version of the message.
            html (str): The HTML version of the message. Optional argument that
                defaults to None.
            images (list): A list of imbedded image file paths for the HTML
                messages. Optional argument that defaults to None.
            attachments (list): A list of file paths for files to attach to the
                email. Optional argument that defaults to None.

        Returns:
            None

        """

        # Overall message object
        msg = MIMEMultipart("mixed")
        msg["Subject"] = subject
        msg["From"] = self.email
        msg["To"] = recipient

        # Create the body
        if (text is not None) or (html is not None):
            msg_body = MIMEMultipart("alternative")

            # If there is a plain text part, attach it
            if text is not None:
                msg_body.attach(MIMEText(text, "plain"))

            # If there is an html part, attach it
            if html is not None:
                # Create a new multipart section
                msg_html = MIMEMultipart("related")
                # Attach the html text
                msg_html.attach(MIMEText(html, "html"))

                # If there are images, include them
                if images is None:
                    images = []
                for i in range(len(images)):
                    # Open the image, read it, and name it so that it can be
                    # referenced by name in the html as:
                    # <img src="cid:image[i]">
                    # where [i] is the index of the image in images
                    with open(images[i], 'rb') as fp:
                        img_type = images[i].split('.')[-1]
                        img = MIMEImage(fp.read(), _subtype=img_type)
                        img.add_header('Content-ID', "<image{}>".format(i))
                    # Attach the image to the html part
                    msg_html.attach(img)

                # Attach the html section to the alternative section
                msg_body.attach(msg_html)

            # Attach the alternative section to the message
            msg.attach(msg_body)

        # Attach each attachment
        for file in attachments or []:
            # Open the file
            with open(file, "rb") as f:
                # Read in the file, and give set the header
                part = MIMEApplication(f.read())
                part.add_header('Content-Disposition',
                        "attachment; filename={}".format(os.path.basename(file)))
                # Attach the attachment to the message
                msg.attach(part)

        # Send the email
        self.server.sendmail(self.email, recipient, msg.as_string())




"""email_processing: Optional processing methods to be used with EmailListener.listen().

Example:

    # Create the EmailListener
    email = "example@gmail.com"
    password = "badpassword"
    folder = "Inbox"
    attachment_dir = "/path/to/attachments"
    el = EmailListener(email, password, folder, attachment_dir)

    # Pass to the listen() function
    timeout = 5
    el.listen(timeout, process_func=write_txt_file)

"""

# Imports from other packages
import json
import os
# Imports from this package
from email_listener.email_responder import EmailResponder


def write_txt_file(email_listener, msg_dict):
    """Write the email message data returned from scrape to text files.

    Args:
        email_listener (EmailListener): The EmailListener object this function
            is used with.
        msg_dict (dict): The dictionary of email message data returned by the
            scraping function.

    Returns:
        A list of file paths of files that were created and written to.

    """

    # List of files to be returned
    file_list = []
    # For each key, create a file and ensure it doesn't exist
    for key in msg_dict.keys():
        file_path = os.path.join(email_listener.attachment_dir, "{}.txt".format(key))
        if os.path.exists(file_path):
            print("File has already been created.")
            continue

        # Open the file
        with open(file_path, "w+") as file:
            # Write the subject to the file
            file.write("Subject\n\n{}\n\n\n".format(msg_dict[key].get("Subject")))
            # Write each section to the file
            for key2 in msg_dict[key].keys():
                val = msg_dict[key][key2]
                # The subject is already writte, skip it
                if (key2 == "Subject"):
                    continue
                # The attachment list will look weird if written as a python list
                if (key2 == "attachments"):
                    files = ""
                    for attachment in val:
                        files = "{}{}\n".format(files, attachment)
                    val = files

                file.write("{}\n\n{}\n\n\n".format(key2, val.strip()))
        # Add the file name to the return list
        file_list.append(file_path)

    return file_list


def send_basic_reply(email_listener, msg_dict):
    """Write the messages to files, and then send a simple automated reply.

    Args:
        email_listener (EmailListener): The EmailListener object this function
            is used with.
        msg_dict (dict): The dictionary of email message data returned by the
            scraping function.

    Returns:
        A list of file paths of files that were created and written to.

    """

    # Write the email messages to files for use later
    file_list = write_txt_file(email_listener, msg_dict)

    er = EmailResponder(email_listener.email, email_listener.app_password)
    er.login()

    # Create the automated response
    subject = "Thank you!"
    message = "Thank you for your email, your request is being processed."

    # For each email
    for key in msg_dict.keys():
        # Split the key up to remove the email uid
        sender_email_parts = key.split('_')
        sender_email = "_".join(sender_email_parts[1:])
        # Send the email
        er.send_singlepart_msg(sender_email, subject, message)

    er.logout()

    return file_list


def write_json_file(email_listener, msg_dict):
    """Write the email message data returned from scrape to json files.

    Args:
        email_listener (EmailListener): The EmailListener object this function
            is used with.
        msg_dict (dict): The dictionary of email message data returned by the
            scraping function.

    Returns:
        A list of file paths of files that were created and written to.

    """

    # List of files to be returned
    file_list = []
    # For each key, create a file and ensure it doesn't exist
    for key in msg_dict.keys():
        file_path = os.path.join(email_listener.attachment_dir, "{}.json".format(key))
        if os.path.exists(file_path):
            print("File has already been created.")
            continue

        # Convert the returned dict to json
        json_obj = json.dumps(msg_dict[key], indent = 4)

        # Open the file
        with open(file_path, "w+") as file:
            # Write the json object to the file
            file.write(json_obj)
        # Add the file name to the return list
        file_list.append(file_path)

    return file_list




"""Helper functions for email_listener.

Examples:

    # Calculate the timeout in 5 minutes
    timeout = calc_timeout(5)

    # Calculate the timeout to be at 1:30pm
    timeout = calc_timeout([13, 30])

    # Get the current time for timeout comparison
    time = get_time()

"""

import datetime


def calc_timeout(timeout):
    """Calculate the time when a timeout should occur in seconds since epoch.

    Args:
        timeout (int or list): Either an integer representing the number of
            minutes to timeout in, or a list, formatted as [hour, minute] of
            the local time to timeout at.

    Returns:
        The timeout time in number of sections since epoch.

    """

    # Get datetime object for the current time
    t = datetime.datetime.now()

    # Calculate time offset based on whether timeout is a
    # list or an int of minutes
    if type(timeout) is list:
        hr = (timeout[0] - t.hour) % 24
        mi = (timeout[1] - t.minute) % 60
        sec = 0 - t.second

        if (timeout[1] - t.minute) < 0:
            hr -= 1
            hr = hr % 24
    elif type(timeout) is int:
        # Convert input minutes to hours and minutes
        hr = timeout // 60
        mi = timeout % 60
        sec = 0
    else:
        # Input isn't an int or list, so it is invalid
        err = ( "timeout must be either a list in the format [hours, minutes] "
                "or an integer representing minutes"
        )
        raise ValueError(err)

    # Calculate the change in time between now and the timeout
    t_delta = datetime.timedelta(seconds=sec, minutes=mi, hours=hr)
    # Calculate the timeout in seconds since epoch
    t_out = (t + t_delta).timestamp()

    # Return the timeout
    return t_out


def get_time():
    """Get the current time in seconds since epoch.

    Args:
        None

    Returns:
        The current time in seconds since epoch.

    """

    return datetime.datetime.now().timestamp()




"""email_listener: Listen in an email folder and process incoming emails.

Example:

    # Create the listener
    listener = EmailListener("example@email.com", "badpassword", "Inbox", "./files/")
    # Log the listener into the IMAP server
    listener.login()
    # Scrape emails from the folder without moving them
    listener.scrape()
    # Scrape emails from the folder, and move them to the "email_listener" folder
    listener.scrape("email_listener")
    # Listen in the folder for 5 minutes, without moving the emails, and not
    # calling any process function on the emails.
    listener.listen(5)
    # Listen in the folder until 1:30pm, moving each new email to the "email_listener"
    # folder, and calling the processing function 'send_reply()'
    listener.listen([13, 30], "email_listener", send_reply)
    # Log the listener out of the IMAP server
    listener.logout()

"""

# Imports from other packages
import email
import html2text
from imapclient import IMAPClient, SEEN
import os
# Imports from this package
from .helpers import (
    calc_timeout,
    get_time,
)
from .email_processing import write_txt_file


class EmailListener:
    """EmailListener object for listening to an email folder and processing emails.

    Attributes:
        email (str): The email to listen to.
        app_password (str): The password for the email.
        folder (str): The email folder to listen in.
        attachment_dir (str): The file path to the folder to save scraped
            emails and attachments to.
        server (IMAPClient): The IMAP server to log into. Defaults to None.

    """

    def __init__(self, email, app_password, folder, attachment_dir):
        """Initialize an EmailListener instance.

        Args:
            email (str): The email to listen to.
            app_password (str): The password for the email.
            folder (str): The email folder to listen in.
            attachment_dir (str): The file path to folder to save scraped
                emails and attachments to.

        Returns:
            None

        """

        self.email = email
        self.app_password = app_password
        self.folder = folder
        self.attachment_dir = attachment_dir
        self.server = None


    def login(self):
        """Logs in the EmailListener to the IMAP server.

        Args:
            None

        Returns:
            None

        """

        self.server = IMAPClient('imap.gmail.com')
        self.server.login(self.email, self.app_password)
        self.server.select_folder(self.folder, readonly=False)


    def logout(self):
        """Logs out the EmailListener from the IMAP server.

        Args:
            None

        Returns:
            None

        """

        self.server.logout()
        self.server = None


    def scrape(self, move=None, unread=False, delete=False):
        """Scrape unread emails from the current folder.

        Args:
            move (str): The folder to move scraped emails to. Defaults to None.

        Returns:
            A list of the file paths to each scraped email.

        """

        # Ensure server is connected
        if type(self.server) is not IMAPClient:
            raise ValueError("server attribute must be type IMAPClient")

        # List containing the file paths of each file created for an email message
        msg_dict = {}

        # Search for unseen messages
        messages = self.server.search("UNSEEN")
        # For each unseen message
        for uid, message_data in self.server.fetch(messages, 'RFC822').items():
            # Get the message
            email_message = email.message_from_bytes(message_data[b'RFC822'])
            # Get who the message is from
            user_raw = email_message.get_all('From', [])
            user_list = email.utils.getaddresses(user_raw)
            if len(user_list[0]) == 1:
                user = user_list[0][0]
            elif len(user_list[0]) == 2:
                user = user_list[0][1]
            else:
                user = "UnknownEmail"

            # Generate the dict key for this email
            key = "{}_{}".format(uid, user)
            # Generate the value dictionary to be filled later
            val_dict = {}

            # Display notice
            print("PROCESSING: Email UID = {} from {}".format(uid, user))

            # Add the subject
            val_dict["Subject"] = email_message.get('Subject').strip()

            # If the email has multiple parts
            if email_message.is_multipart():
                # For each part
                for part in email_message.walk():
                    # If the part is multipart, pass
                    if part.get_content_maintype() == 'multipart':
                        continue

                    # If the part is an attachment
                    file_name = part.get_filename()
                    if bool(file_name):
                        # Generate file path
                        file_path = os.path.join(self.attachment_dir, file_name)
                        with open(file_path, 'wb') as file:
                            file.write(part.get_payload(decode=True))
                        attachment_list = val_dict.get("attachments")
                        if attachment_list is None:
                            val_dict["attachments"] = ["{}".format(file_path)]
                        else:
                            attachment_list.append("{}".format(file_path))
                            val_dict["attachments"] = attachment_list

                    # If the part is html text
                    elif part.get_content_type() == 'text/html':
                        # Convert the body from html to plain text
                        val_dict["Plain_HTML"] = html2text.html2text(
                                part.get_payload())
                        val_dict["HTML"] = part.get_payload()

                    # If the part is plain text
                    elif part.get_content_type() == 'text/plain':
                        # Get the body
                        val_dict["Plain_Text"] = part.get_payload()

            # If the message isn't multipart
            else:
                val_dict["Plain_Text"] = email_message.get_payload()

            msg_dict[key] = val_dict

            # If the message should be marked as unread
            if bool(unread):
                self.server.remove_flags(uid, [SEEN])

            # If a move folder is specified
            if move is not None:
                try:
                    # Move the message to another folder
                    self.server.move(uid, move)
                except:
                    # Create the folder and move the message to the folder
                    self.server.create_folder(move)
                    self.server.move(uid, move)
            elif bool(delete):
                # Add delete statement
                self.server.set_gmail_labels(uid, "\\Trash")

        return msg_dict


    def listen(self, timeout, process_func=write_txt_file, move=None,
               unread=False, delete=False):
        """Listen in an email folder for incoming emails, and process them.

        Args:
            timeout (int or list): Either an integer representing the number
                of minutes to timeout in, or a list, formatted as [hour, minute]
                of the local time to timeout at.
            move (str): The name of the folder to move processed emails to. If
                None, the emails are not moved or marked as read. Defaults to
                None.
            process_func (function): A function called to further process the
                emails. The function must take only the list of file paths
                returned by the scrape function as an argument. Defaults to the
                example function write_txt_file in the email_processing module.

        Returns:
            None

        """

        # Ensure server is connected
        if type(self.server) is not IMAPClient:
            raise ValueError("server attribute must be type IMAPClient")

        # Get the timeout value
        outer_timeout = calc_timeout(timeout)        

        # Run until the timeout is reached
        while (get_time() < outer_timeout):
            # Start idling
            self.server.idle()
            print("Connection is now in IDLE mode.")
            # Set idle timeout to 5 minutes
            inner_timeout = get_time() + 60*5
            # Until idle times out
            while (get_time() < inner_timeout):
                # Check for a new response every 30 seconds
                responses = self.server.idle_check(timeout=30)
                print("Server sent:", responses if responses else "nothing")
                # If there is a response
                if (responses):
                    # Suspend the idling
                    self.server.idle_done()
                    # Process the new emails
                    msgs = self.scrape(move=move, unread=unread, delete=delete)
                    # Run the process function
                    process_func(self, msgs)
                    # Restart idling
                    self.server.idle()
            # Stop idling
            self.server.idle_done()
        return




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='email_listener',
    version='1.1',
    description='Listen in an email folder and process incoming emails.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Communications :: Email',
    ],
    url='https://github.com/njdreikosen/email_listener',
    author='Noah Dreikosen',
    author_email='ndreikosen@gmail.com',
    license='GNU GPLv3',
    packages=['email_listener'],
    install_requires=[
        'datetime',
        'html2text',
        'imapclient',
        'pytest',
    ],
    zip_safe=False)

